import signal
from functools import wraps, partial
from collections import defaultdict

from curio import SignalEvent
from curio import run, spawn, tcp_server
from autoroutes import Routes

from trinket import lifecycle
from trinket.handler import request_handler
from trinket.request import Request
from trinket.http import HTTPStatus, HTTPError
from trinket.websockets import Websocket


Goodbye = SignalEvent(signal.SIGINT, signal.SIGTERM)


class Trinket:

    __slots__ = ('hooks', 'routes', 'websockets')

    def __init__(self):
        self.routes = Routes()
        self.websockets = set()
        self.hooks = defaultdict(list)

    async def lookup(self, request: Request):
        payload, params = self.routes.match(request.path)

        if not payload:
            raise HTTPError(HTTPStatus.NOT_FOUND, request.path)

        # Uppercased in order to only consider HTTP verbs.
        handler = payload.get(request.method.upper(), None)
        if handler is None:
            raise HTTPError(HTTPStatus.METHOD_NOT_ALLOWED)

        # We check if the route is for a websocket handler.
        # If it is, we make sure we were asked for an upgrade.
        if payload.get('websocket', False) and not request.upgrade:
            raise HTTPError(
                HTTPStatus.UPGRADE_REQUIRED,
                'This is a websocket endpoint, please upgrade.')

        return handler, params

    @lifecycle.handler_events
    async def __call__(self, request: Request):
        handler, params = await self.lookup(request)
        return await handler(request, **params)

    def route(self, path: str, methods: list=None, **extras: dict):
        if methods is None:
            methods = ['GET']

        def wrapper(func):
            payload = {method: func for method in methods}
            payload.update(extras)
            self.routes.add(path, **payload)
            return func

        return wrapper

    def websocket(self, path: str, **extras: dict):

        def wrapper(func):
            @wraps(func)
            async def websocket_handler(request, **params):
                websocket = Websocket(request.socket)
                try:
                    await websocket.upgrade(request)
                    self.websockets.add(websocket)
                    task = await spawn(func, request, websocket, **params)
                    await websocket.flow(task)
                finally:
                    self.websockets.discard(websocket)

            payload = {'GET': websocket_handler, 'websocket': True}
            payload.update(extras)
            self.routes.add(path, **payload)
            return func

        return wrapper

    def listen(self, name: str):
        def wrapper(func):
            self.hooks[name].append(func)
        return wrapper

    async def notify(self, name: str, *args, **kwargs):
        if name in self.hooks:
            for hook in self.hooks[name]:
                result = await hook(*args, **kwargs)
                if result is not None:
                    # Allows to shortcut the chain.
                    return result
        return None

    @lifecycle.server_events
    async def serve(self, host, port):
        print('Trinket serving on {}:{}'.format(host, port))
        handler = partial(request_handler, self)
        server = await spawn(tcp_server(host, port, handler))
        await Goodbye.wait()
        print('Server is shutting down.')
        print('Please wait. The remaining tasks are being terminated.')
        await server.cancel()

    def start(self, host='127.0.0.1', port=5000, debug=True):
        run(self.serve, host, port, with_monitor=debug)
        print('Trinket is crumbling away...')



from io import BytesIO
from multifruits import Parser, extract_filename, parse_content_disposition
from trinket.http import Form, Files, HTTPError, HTTPStatus


class Multipart:
    """Responsible of the parsing of multipart encoded `request.body`.
    """

    __slots__ = (
        'form',
        'files',
        '_parser',
        '_current',
        '_current_headers',
        '_current_params')

    def __init__(self, content_type: str):
        self._parser = Parser(self, content_type.encode())
        self.form = Form()
        self.files = Files()

    def feed_data(self, data: bytes):
        self._parser.feed_data(data)

    def on_part_begin(self):
        self._current_headers = {}

    def on_header(self, field: bytes, value: bytes):
        self._current_headers[field] = value

    def on_headers_complete(self):
        disposition_type, params = parse_content_disposition(
            self._current_headers.get(b'Content-Disposition'))
        if not disposition_type:
            return
        self._current_params = params
        if b'Content-Type' in self._current_headers:
            self._current = BytesIO()
            self._current.filename = extract_filename(params)
            self._current.content_type = self._current_headers[b'Content-Type']
            self._current.params = params
        else:
            self._current = ''

    def on_data(self, data: bytes):
        if b'Content-Type' in self._current_headers:
            self._current.write(data)
        else:
            self._current += data.decode()

    def on_part_complete(self):
        name = self._current_params.get(b'name', b'').decode()
        if b'Content-Type' in self._current_headers:
            if name not in self.files:
                self.files[name] = []
            self._current.seek(0)
            self.files[name].append(self._current)
        else:
            if name not in self.form:
                self.form[name] = []
            self.form[name].append(self._current)
        self._current = None


def read_multipart(content_type):
    parser = Multipart(content_type)
    while True:
        chunk = yield
        if not chunk:
            yield parser.form, parser.files
            break
        try:
            parser.feed_data(chunk)
        except ValueError:
            raise HTTPError(
                HTTPStatus.BAD_REQUEST,
                'Unparsable multipart body')



from functools import wraps


def handler_events(func):
    @wraps(func)
    async def dispatch(app, request, *args, **kwargs):
        response = await app.notify('request', request)
        if response is None:
            response = await func(app, request, *args, **kwargs)
        if response is not None:
            await app.notify('response', request, response)
        return response
    return dispatch


def server_events(func):
    @wraps(func)
    async def dispatch(app, *args, **kwargs):
        await app.notify('startup')
        await func(app, *args, **kwargs)
        await app.notify('shutdown')
    return dispatch



from biscuits import parse
from trinket.http import HTTPStatus, HTTPError, Query
from trinket.parsers import CONTENT_TYPES_PARSERS
from httptools import HttpParserUpgrade, HttpParserError, HttpRequestParser
from httptools.parser.errors import HttpParserInvalidMethodError
from httptools import parse_url
from urllib.parse import parse_qs, unquote


class Channel:

    __slots__ = (
        'parser',
        'request',
        'complete',
        'headers_complete',
        'socket',
        'reader',
    )

    def __init__(self, socket):
        self.complete = False
        self.headers_complete = False
        self.parser = HttpRequestParser(self)
        self.request = None
        self.socket = socket
        self.reader = self._reader()

    def data_received(self, data: bytes):
        try:
            self.parser.feed_data(data)
        except HttpParserUpgrade:
            self.request.upgrade = True
        except (HttpParserError, HttpParserInvalidMethodError) as exc:
            raise HTTPError(
                HTTPStatus.BAD_REQUEST, 'Unparsable request.')

    async def read(self, parse: bool=True) -> bytes:
        data = await self.socket.recv(8096)
        if data:
            if parse:
                self.data_received(data)
            return data

    async def _reader(self) -> bytes:
        while not self.complete:
            data = await self.read()
            if not data:
                break
            yield data

    async def _drainer(self) -> bytes:
        while True:
            data = await self.read(parse=False)
            if not data:
                break
            yield data

    def on_header(self, name: bytes, value: bytes):
        value = value.decode()
        if value:
            name = name.decode().title()
            if name in self.request.headers:
                self.request.headers[name] += ', {}'.format(value)
            else:
                self.request.headers[name] = value

    def on_body(self, data: bytes):
        self.request.body += data

    def on_message_begin(self):
        self.complete = False
        self.request = Request(self.socket, self.reader)

    def on_message_complete(self):
        self.complete = True

    def on_url(self, url: bytes):
        self.request.url = url
        parsed = parse_url(url)
        self.request.path = unquote(parsed.path.decode())
        self.request.query_string = (parsed.query or b'').decode()

    def on_headers_complete(self):
        self.request.keep_alive = self.parser.should_keep_alive()
        self.request.method = self.parser.get_method().decode().upper()
        self.headers_complete = True

    async def __aiter__(self):
        keep_alive = True
        while keep_alive:
            data = await self.read()
            if data is None:
                break
            if self.headers_complete:
                yield self.request
                keep_alive = self.request.keep_alive
                if keep_alive:
                    if not self.complete:
                        await self.reader.aclose()
                        # We drain if there's an uncomplete request.
                        async for _ in self._drainer():
                            pass
                    self.request = None
                    self.complete = False
                    self.headers_complete = False
                    self.reader = self._reader()


class Request(dict):

    __slots__ = (
        '_cookies',
        '_query',
        '_reader',
        'body',
        'files',
        'form',
        'headers',
        'keep_alive',
        'method',
        'path',
        'query_string',
        'socket',
        'upgrade',
        'url'
    )

    def __init__(self, socket, reader, **headers):
        self._cookies = None
        self._query = None
        self._reader = reader
        self.body = b''
        self.files = None
        self.form = None
        self.headers = headers
        self.keep_alive = False
        self.method = None
        self.path = None
        self.query_string = None
        self.socket = socket
        self.upgrade = False
        self.url = None

    @property
    async def raw_body(self):
        async for data in self._reader:
            self.body += data
        return self.body

    async def parse_body(self):
        disposition = self.content_type.split(';', 1)[0]
        parser_type = CONTENT_TYPES_PARSERS.get(disposition)
        if parser_type is None:
            raise NotImplementedError(f"Don't know how to parse {disposition}")
        content_parser = parser_type(self.content_type)
        next(content_parser)
        if self.body:
            content_parser.send(self.body)

        async for data in self._reader:
            content_parser.send(data)

        self.form, self.files = next(content_parser)
        content_parser.close()

    @property
    def cookies(self):
        if self._cookies is None:
            self._cookies = parse(self.headers.get('Cookie', b''))
        return self._cookies

    @property
    def query(self) -> Query:
        if self._query is None:
            parsed_qs = parse_qs(self.query_string, keep_blank_values=True)
            self._query = Query(parsed_qs)
        return self._query

    @property
    def content_type(self) -> bytes:
        return self.headers.get('Content-Type', b'')

    @property
    def host(self) -> bytes:
        return self.headers.get('Host', b'')



import http.client
import mimetypes
from contextlib import asynccontextmanager

import pytest
import curio

from io import BytesIO
from uuid import uuid4
from urllib.parse import urlencode

from trinket.app import Trinket
from trinket.websockets import WebsocketPrototype
from wsproto.connection import WSConnection, CLIENT
from wsproto.events import ConnectionEstablished


def encode_multipart(data, charset='utf-8'):
    # Ported from Werkzeug testing.
    boundary = '---------------Boundary%s' % uuid4().hex
    body = BytesIO()

    def write(string):
        body.write(string.encode(charset))

    if isinstance(data, dict):
        data = data.items()

    for key, values in data:
        if not isinstance(values, (list, tuple)):
            values = [values]
        for value in values:
            write('--%s\r\nContent-Disposition: form-data; name="%s"' %
                  (boundary, key))
            reader = getattr(value, 'read', None)
            if reader is not None:
                filename = getattr(value, 'filename',
                                   getattr(value, 'name', None))
                content_type = getattr(value, 'content_type', None)
                if content_type is None:
                    content_type = filename and \
                        mimetypes.guess_type(filename)[0] or \
                        'application/octet-stream'
                if filename is not None:
                    write('; filename="%s"\r\n' % filename)
                else:
                    write('\r\n')
                write('Content-Type: %s\r\n\r\n' % content_type)
                while 1:
                    chunk = reader(16384)
                    if not chunk:
                        break
                    body.write(chunk)
            else:
                if not isinstance(value, str):
                    value = str(value)
                else:
                    value = value.encode(charset)
                write('\r\n\r\n')
                body.write(value)
            write('\r\n')
    write('--%s--\r\n' % boundary)

    body.seek(0)
    content_type = 'multipart/form-data; boundary=%s' % boundary
    return body.read(), content_type


class RequestForger:

    @staticmethod
    def prepare_files(files):
        body = {}
        if isinstance(files, dict):
            files = files.items()
        for key, els in files:
            if not els:
                continue
            if not isinstance(els, (list, tuple)):
                # Allow passing a file instance.
                els = [els]
            file_ = els[0]
            if isinstance(file_, str):
                file_ = file_.encode()
            if isinstance(file_, bytes):
                file_ = BytesIO(file_)
            if len(els) > 1:
                file_.name = els[1]
            if len(els) > 2:
                file_.charset = els[2]
            body[key] = file_
        return 'multipart/form-data', body

    @staticmethod
    def encode_body(body, headers):
        if not body or isinstance(body, (str, bytes)):
            return body, headers

        content_type = headers.get('Content-Type')
        if content_type:
            if 'application/x-www-form-urlencoded' in content_type:
                body = urlencode(body)
            elif 'application/json' in content_type:
                body = json.dumps(body)
            elif 'multipart/form-data' in content_type:
                body, headers['Content-Type'] = encode_multipart(body)
            else:
                raise NotImplementedError('Content-Type not supported')
        return body, headers

    @classmethod
    def forge(cls, method, path, body, headers=None, content_type=None):
        headers = headers or {}
        if content_type:
            headers['Content-Type'] = content_type
        body, headers = cls.encode_body(body, headers)
        if isinstance(body, str):
            body = body.encode()
        if body and 'Content-Length' not in headers:
            headers['Content-Length'] = len(body)
        headers = '\r\n'.join('{}: {}'.format(*h) for h in headers.items())
        data = b'%b %b HTTP/1.1\r\n%b\r\n\r\n%b' % (
            method.encode(), path.encode(), headers.encode(), body or b'')
        return data

    @classmethod
    def post(cls, path, body=b'', headers=None, content_type=None, files=None):

        if not body:
            body = {}

        if files is not None:
            content_type, contents = cls.prepare_files(files)
            body.update(contents)
    
        return cls.forge('POST', path, body, headers, content_type)


class Websocket(WebsocketPrototype):

    def __init__(self, host, path):
        super().__init__()
        self.host = host
        self.socket = curio.socket.socket(
            curio.socket.AF_INET, curio.socket.SOCK_STREAM)
        self.protocol = WSConnection(CLIENT, host, path)

    async def connect(self, host, port):
        await self.socket.connect((host, port))
        data = self.protocol.bytes_to_send()
        await self.socket.sendall(data)
        upgrade_response = await self.socket.recv(8096)
        self.protocol.receive_bytes(upgrade_response)
        event = next(self.protocol.events())
        if isinstance(event, ConnectionEstablished):
            print('WebSocket negotiation complete')
        else:
            raise Exception('Expected ConnectionEstablished event!')


class LiveClient:

    def __init__(self, app: Trinket, port: int=42000):
        self.app = app
        self.port = port
        self.task = None

    async def __aenter__(self):
        self.task = await curio.spawn(self.app.serve, '', self.port)
        await curio.sleep(0.1)

    async def __aexit__(self, *args, **kwargs):
        await self.task.cancel()

    @asynccontextmanager
    async def query(self, method, uri, headers: dict=None, body=None):
        conn = http.client.HTTPConnection('localhost', self.port)

        def execute(method, uri, headers, body):
            if headers is None:
                headers = {}

            conn.request(method, uri, headers=headers, body=body)
            response = conn.getresponse()
            return response

        yield await curio.run_in_thread(execute, method, uri, headers, body)
        conn.close()

    @asynccontextmanager
    async def websocket(self, resource):
        ws = Websocket(f'localhost:{self.port}', resource)
        await ws.connect('localhost', self.port)
        task = await curio.spawn(ws.flow)
        yield ws
        try:
            await ws.socket.shutdown(curio.socket.SHUT_RDWR)
        except curio.socket.error:
            # socket was killed.
            pass
        await task.join()


@pytest.fixture
def app():
    return Trinket()


@pytest.fixture
def client(app):
    return LiveClient(app)



from trinket.http import Form, Files, HTTPError, HTTPStatus
from urllib.parse import parse_qs


def read_urlencoded(content_type):
    data = b''
    while True:
        chunk = yield
        if not chunk:
            try:
                parsed = parse_qs(
                    data, keep_blank_values=True, strict_parsing=True)
            except ValueError:
                raise HTTPError(
                    HTTPStatus.BAD_REQUEST,
                    'Unparsable urlencoded body')
            yield Form(parsed), Files()
            raise StopIteration
        data += chunk



from http import HTTPStatus
from io import BytesIO
from typing import TypeVar

from biscuits import Cookie
from multifruits import Parser, extract_filename, parse_content_disposition


HTTPCode = TypeVar('HTTPCode', HTTPStatus, int)


class HTTPError(Exception):
    """Exception meant to be raised when an error is occurring.

    E.g.:
        Within your view `raise HTTPError(HTTPStatus.BAD_REQUEST)` will
        direcly return a 400 HTTP status code with descriptive content.
    """

    __slots__ = ('status', 'message')

    def __init__(self, http_code: HTTPCode, message=None):
        # Idempotent if `http_code` is already an `HTTPStatus` instance.
        self.status = HTTPStatus(http_code)
        if isinstance(message, str):
            message = message.encode()
        self.message = message or self.status.phrase.encode()

    def __bytes__(self):
        return b'HTTP/1.1 %a %b\r\nContent-Length: %a\r\n\r\n%b' % (
            self.status.value, self.status.phrase.encode(),
            len(self.message), self.message)


class Multidict(dict):
    """Data structure to deal with several values for the same key.

    Useful for query string parameters or form-like POSTed ones.
    """

    def get(self, key: str, default=...):
        return self.list(key, [default])[0]

    def list(self, key: str, default=...):
        try:
            return self[key]
        except KeyError:
            if default is ... or default == [...]:
                raise HTTPError(HTTPStatus.BAD_REQUEST,
                                "Missing '{}' key".format(key))
            return default


class Query(Multidict):
    """Allow to access casted GET parameters from `request.query`.

    E.g.:
        `request.query.int('weight', 0)` will return an integer or zero.
    """

    TRUE_STRINGS = ('t', 'true', 'yes', '1', 'on')
    FALSE_STRINGS = ('f', 'false', 'no', '0', 'off')
    NONE_STRINGS = ('n', 'none', 'null')

    def bool(self, key: str, default=...):
        value = self.get(key, default)
        if value in (True, False, None):
            return value
        value = value.lower()
        if value in self.TRUE_STRINGS:
            return True
        elif value in self.FALSE_STRINGS:
            return False
        elif value in self.NONE_STRINGS:
            return None
        raise HTTPError(
            HTTPStatus.BAD_REQUEST,
            "Wrong boolean value for '{}={}'".format(key, self.get(key)))

    def int(self, key: str, default=...):
        try:
            return int(self.get(key, default))
        except ValueError:
            raise HTTPError(HTTPStatus.BAD_REQUEST,
                            "Key '{}' must be castable to int".format(key))

    def float(self, key: str, default=...):
        try:
            return float(self.get(key, default))
        except ValueError:
            raise HTTPError(HTTPStatus.BAD_REQUEST,
                            "Key '{}' must be castable to float".format(key))


class Form(Query):
    """Allow to access casted POST parameters from `request.body`."""


class Files(Multidict):
    """Allow to access POSTed files from `request.body`."""


class Multipart:
    """Responsible of the parsing of multipart encoded `request.body`."""

    __slots__ = ('app', 'form', 'files', '_parser', '_current',
                 '_current_headers', '_current_params')

    def __init__(self, app):
        self.app = app

    def initialize(self, content_type: str):
        self._parser = Parser(self, content_type.encode())
        self.form = self.app.Form()
        self.files = self.app.Files()
        return self.form, self.files

    def feed_data(self, data: bytes):
        self._parser.feed_data(data)

    def on_part_begin(self):
        self._current_headers = {}

    def on_header(self, field: bytes, value: bytes):
        self._current_headers[field] = value

    def on_headers_complete(self):
        disposition_type, params = parse_content_disposition(
            self._current_headers.get(b'Content-Disposition'))
        if not disposition_type:
            return
        self._current_params = params
        if b'Content-Type' in self._current_headers:
            self._current = BytesIO()
            self._current.filename = extract_filename(params)
            self._current.content_type = self._current_headers[b'Content-Type']
            self._current.params = params
        else:
            self._current = ''

    def on_data(self, data: bytes):
        if b'Content-Type' in self._current_headers:
            self._current.write(data)
        else:
            self._current += data.decode()

    def on_part_complete(self):
        name = self._current_params.get(b'name', b'').decode()
        if b'Content-Type' in self._current_headers:
            if name not in self.files:
                self.files[name] = []
            self._current.seek(0)
            self.files[name].append(self._current)
        else:
            if name not in self.form:
                self.form[name] = []
            self.form[name].append(self._current)
        self._current = None


class Cookies(dict):
    """A Cookies management class, built on top of biscuits."""

    def set(self, name, *args, **kwargs):
        self[name] = Cookie(name, *args, **kwargs)



import socket
from trinket.request import Channel
from trinket.response import response_handler
from trinket.http import HTTPError


async def request_handler(app, client, addr):
    async with client:
        try:
            async for request in Channel(client):
                response = await app(request)
                if response is None:
                    break
                await response_handler(client, response)
        except HTTPError as exc:
            await client.sendall(bytes(exc))
        except (ConnectionResetError, BrokenPipeError, socket.timeout):
            # The client disconnected or the network is suddenly
            # unreachable.
            pass



from abc import ABC
from curio import socket, Queue, Event, TaskGroup
from trinket.http import HTTPStatus, HTTPError
from wsproto.connection import WSConnection, SERVER
from wsproto import events
from wsproto.extensions import PerMessageDeflate


DATA_TYPES = (events.TextReceived, events.BytesReceived)


class WebsocketClosedError(Exception):
    pass


class WebsocketPrototype(ABC):

    socket = None
    protocol = None

    def __init__(self):
        self.outgoing = Queue()
        self.incoming = Queue()
        self.closure = None
        self.closing = Event()

    @property
    def closed(self):
        return self.closing.is_set()

    async def send(self, data):
        if self.closed:
            raise WebsocketClosedError()
        await self.outgoing.put(data)

    async def recv(self):
        if not self.closed:
            async with TaskGroup(wait=any) as g:
                receiver = await g.spawn(self.incoming.get)
                await g.spawn(self.closing.wait)
            if g.completed is receiver:
                return receiver.result

    async def __aiter__(self):
        async for msg in self.incoming:
            yield msg

    async def close(self, code=1000, reason='Closed.'):
        await self.outgoing.put(events.ConnectionClosed(code, reason))

    async def _handle_incoming(self):
        while not self.closed:
            data = await self.socket.recv(4096)
            self.protocol.receive_bytes(data)
            if not data:
                await self.closing.set()
            for event in self.protocol.events():
                cl = event.__class__
                if cl in DATA_TYPES:
                    await self.incoming.put(event.data)
                elif cl is events.ConnectionClosed:
                    # The client has closed the connection gracefully.
                    self.closure = event
                    await self.closing.set()

    async def _handle_outgoing(self):
        async for data in self.outgoing:
            if data is None:
                await self.closing.set()
            elif isinstance(data, events.ConnectionClosed):
                self.protocol.close(code=data.code, reason=data.reason)
                self.closure = data
                await self.closing.set()
            else:
                self.protocol.send_data(data)
            try:
                await self.socket.sendall(self.protocol.bytes_to_send())
            except socket.error:
                await self.closing.set()
            if self.closed:
                return

    async def flow(self, *tasks):
        async with TaskGroup(tasks=tasks) as ws:
            incoming = await ws.spawn(self._handle_incoming)
            outgoing = await ws.spawn(self._handle_outgoing)
            finished = await ws.next_done()
            if finished is incoming:
                await self.outgoing.put(None)
                await outgoing.join()
            elif finished in tasks:
                # Task is finished.
                # We ask for the outgoing to finish
                if finished.exception:
                    await self.close(1011, 'Task died prematurely.')
                else:
                    await self.close()
                await outgoing.join()


class Websocket(WebsocketPrototype):
    """Server-side websocket running a handler parallel to the I/O.
    """
    def __init__(self, socket):
        super().__init__()
        self.socket = socket
        self.protocol = WSConnection(SERVER, extensions=[PerMessageDeflate()])

    async def upgrade(self, request):
        data = '{} {} HTTP/1.1\r\n'.format(
            request.method, request.url)
        data += '\r\n'.join(
            ('{}: {}'.format(k, v)
             for k, v in request.headers.items())) + '\r\n\r\n'

        data = data.encode()
        self.protocol.receive_bytes(data)
        event = next(self.protocol.events())
        if not isinstance(event, events.ConnectionRequested):
            raise HTTPError(HTTPStatus.BAD_REQUEST)

        self.protocol.accept(event)
        data = self.protocol.bytes_to_send()
        await self.socket.sendall(data)



import logging


def logger(app, level=logging.DEBUG):

    logger = logging.getLogger('trinket')
    logger.setLevel(level)
    handler = logging.StreamHandler()

    @app.listen('request')
    async def log_request(request):
        logger.info('%s %s', request.method, request.url.decode())

    @app.listen('startup')
    async def startup():
        logger.addHandler(handler)

    @app.listen('shutdown')
    async def shutdown():
        logger.removeHandler(handler)

    return app



try:
    # In case you use json heavily, we recommend installing
    # https://pypi.org/project/python-rapidjson for better performances.
    import rapidjson as json
except ImportError:
    import json as json

import curio
from collections.abc import AsyncGenerator
from trinket.http import HTTPCode, HTTPStatus, Cookies


async def file_iterator(path):
    async with curio.aopen(path, 'rb') as reader:
        while True:
            data = await reader.read(4096)
            if not data:
                break
            yield data


async def response_handler(client, response):
    """The bytes representation of the response
    contains a body only if there's no streaming
    In a case of a stream, it only contains headers.
    """
    await client.sendall(bytes(response))

    if response.stream is not None:
        if isinstance(response.stream, AsyncGenerator):
            async with curio.meta.finalize(response.stream):
                async for data in response.stream:
                    await client.sendall(
                        b"%x\r\n%b\r\n" % (len(data), data))
        else:
            for data in response.stream:
                await client.sendall(
                    b"%x\r\n%b\r\n" % (len(data), data))

        await client.sendall(b'0\r\n\r\n')


class Response:
    """A container for `status`, `headers` and `body`."""

    __slots__ = (
        'headers', 'body', 'bodyless', '_cookies', '_status', 'stream')

    BODYLESS_METHODS = frozenset(('HEAD', 'CONNECT'))
    BODYLESS_STATUSES = frozenset((
        HTTPStatus.CONTINUE, HTTPStatus.SWITCHING_PROTOCOLS,
        HTTPStatus.PROCESSING, HTTPStatus.NO_CONTENT,
        HTTPStatus.NOT_MODIFIED))

    def __init__(self, status=HTTPStatus.OK, body=b'', headers=None):
        self._cookies = None
        self.status = status
        self.body = body
        if headers is None:
            headers = {}
        self.headers = headers
        self.stream = None

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, http_code: HTTPCode):
        self._status = HTTPStatus(http_code)
        self.bodyless = self._status in self.BODYLESS_STATUSES

    @classmethod
    def json(cls, value):
        body = json.dumps(value)
        headers = {'Content-Type': 'application/json; charset=utf-8'}
        return cls(body=body, headers=headers)

    @classmethod
    def raw(cls, value: bytes):
        headers = {'Content-Type': 'text/plain; charset=utf-8'}
        return cls(body=value, headers=headers)

    @classmethod
    def html(cls, value: bytes):
        headers = {'Content-Type': 'text/html; charset=utf-8'}
        return cls(body=value, headers=headers)

    @classmethod
    def streamer(self, gen, content_type="application/octet-stream"):
        headers = {
            'Content-Type': content_type,
            'Transfer-Encoding': 'chunked',
            'Keep-Alive': 10,
        }
        response = Response(headers=headers)
        response.stream = gen
        return response

    @property
    def cookies(self):
        if self._cookies is None:
            self._cookies = Cookies()
        return self._cookies

    def __bytes__(self):
        response = b'HTTP/1.1 %a %b\r\n' % (
            self.status.value, self.status.phrase.encode())

        if self._cookies:
            # https://tools.ietf.org/html/rfc7230#page-23
            for cookie in self.cookies.values():
                response += b'Set-Cookie: %b\r\n' % str(cookie).encode()

        # https://tools.ietf.org/html/rfc7230#section-3.3.2 :scream:
        for key, value in self.headers.items():
            response += b'%b: %b\r\n' % (key.encode(), str(value).encode())

        if not self.bodyless:
            if not isinstance(self.body, bytes):
                body = str(self.body).encode()
            else:
                body = self.body

            if self.stream is None and 'Content-Length' not in self.headers:
                response += b'Content-Length: %i\r\n' % len(body)

            response += b'\r\n'
            if body and self.stream is None:
                # We don't write the body if there's a stream.
                # It takes precedence
                response += body
        else:
            response += b'\r\n'
        return response



from os import path
from setuptools import setup, find_packages


def read(filename):
    with open(path.join(path.dirname(__file__), filename)) as f:
        return f.read()


install_requires = [
    "autoroutes >= 0.2.0",
    "biscuits >= 0.1.1",
    "httptools >= 0.0.11",
    "multifruits >= 0.1.1",
    "curio >= 0.9",
    "wsproto >= 0.12.0"
    ]

tests_require = [
    "pytest",
    ]

setup(name='trinket',
      version='0.1',
      description="Curio-based web framework",
      long_description="%s\n\n%s" % (
          read('README.md'), read(path.join('docs', 'HISTORY.txt'))),
      keywords="Curio HTTP",
      author="",
      author_email="",
      license="BSD",
      packages=find_packages('src', exclude=['ez_setup']),
      package_dir={'': 'src'},
      include_package_data=True,
      zip_safe=False,
      install_requires=install_requires,
      extras_require={'test': tests_require},
      entry_points={
          'pytest11': ['trinket=trinket.testing'],
      }
      )



from .multipart import read_multipart
from .urlencoded import read_urlencoded


CONTENT_TYPES_PARSERS = {
    'multipart/form-data': read_multipart,
    'application/x-www-form-urlencoded': read_urlencoded,
}


__all__ = ['CONTENT_TYPES_PARSERS']
